# requires requests and cryptography packages.
import json
import time
import base64
import hashlib
import email.utils
import argparse
from collections import OrderedDict

import requests

from cryptography.hazmat.primitives.hashes import SHA256
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from urllib.parse import urlparse

CLIENT_ID = "client_id"
CLIENT_SECRET = "client_secret"
CERTIFICATE_PRIVATE_KEY_FILE_NAME = "key.pem"


def get_common_headers(prepared_req):
    host = urlparse(prepared_req.url).netloc
    date = email.utils.formatdate(time.time(), usegmt=True)
    path = requests.models.RequestEncodingMixin.path_url.fget(prepared_req)
    request_target = "{} {}".format(
        prepared_req.method.lower(),
        path,
    )
    headers = OrderedDict(
        (
            ("(request-target)", request_target),
            ("Host", host),
            ("Date", date),
        )
    )
    return headers


def calculate_request_digest(prepared_req):
    digest = hashlib.sha256(
        prepared_req.body,
    ).digest()
    return "SHA-256=" + base64.b64encode(digest).decode()


def get_encrypted_signature(headers, private_key):
    string_to_sign = "\n".join(
        "{}: {}".format(name.lower(), value) for name, value in headers.items()
    )
    hasher = SHA256()
    raw_signature = private_key.sign(
        string_to_sign.encode(),
        padding=PKCS1v15(),
        algorithm=hasher,
    )
    return base64.b64encode(raw_signature).decode()


def get_signature_header(key_id, headers, private_key):
    encrypted_signature = get_encrypted_signature(headers, private_key)
    return (
        'keyId="{key_id}",'
        'algorithm="rsa-sha256",'
        'headers="{headers}",'
        'signature="{encrypted_sig}"'.format(
            key_id=key_id,
            headers=" ".join(name.lower() for name in headers),
            encrypted_sig=encrypted_signature,
        )
    )


def get_headers_with_content(prepared_req):
    headers = get_common_headers(prepared_req)
    content_type = prepared_req.headers.get("Content-Type", None)
    if content_type is None:
        raise ValueError("Content-Type header is required for requests with content")
    headers["Content-Type"] = content_type
    headers["Digest"] = calculate_request_digest(prepared_req)
    return headers


def get_headers(
    prepared_req,
    private_key,
    holvi_client_id,
    holvi_client_secret,
    user_jwt_token=None,
):
    if prepared_req.method.lower() in ("post", "patch", "put"):
        headers = get_headers_with_content(prepared_req)
    else:
        headers = get_common_headers(prepared_req)
    headers["Signature"] = get_signature_header(
        holvi_client_id,
        headers,
        private_key,
    )
    headers["X-Holvi-Client-Id"] = holvi_client_id
    headers["X-Holvi-Client-Secret"] = holvi_client_secret

    if user_jwt_token:
        headers["Authorization"] = f"Bearer {user_jwt_token}"

    return headers


def make_request(
    method,
    url,
    holvi_client_id,
    holvi_client_secret,
    private_key,
    user_jwt_token=None,
    **kwargs,
):
    prepared_req = requests.Request(method, url, **kwargs).prepare()
    headers = get_headers(
        prepared_req,
        private_key,
        holvi_client_id,
        holvi_client_secret,
        user_jwt_token,
    )
    # (request-target) is used for generation, but it's not a correct header to send
    headers.pop("(request-target)")
    prepared_req.headers.update(headers)
    session = requests.Session()
    return session.send(prepared_req)


def load_private_key():
    with open(CERTIFICATE_PRIVATE_KEY_FILE_NAME, "rb") as f:
        private_key = load_pem_private_key(
            f.read(),
            password=None,
            backend=default_backend(),
        )
    return private_key


# GET Request
def make_get_request(url: str, user_jwt_token: str = None):
    private_key = load_private_key()
    response = make_request(
        method="GET",
        url=url,
        holvi_client_id=CLIENT_ID,
        holvi_client_secret=CLIENT_SECRET,
        private_key=private_key,
        user_jwt_token=user_jwt_token,
        # params={"state": "paid"}, # TODO
    )
    stdout_response(response)


# POST Request
def make_post_request(url: str, payload: dict, user_jwt_token: str = None):
    private_key = load_private_key()

    response = make_request(
        method="POST",
        url=url,
        holvi_client_id=CLIENT_ID,
        holvi_client_secret=CLIENT_SECRET,
        private_key=private_key,
        user_jwt_token=user_jwt_token,
        json=payload,
    )
    stdout_response(response)


def stdout_response(response):
    print("Status code:", response.status_code)
    print("Response content: \n")
    print(json.dumps(json.loads(response.text), indent=2), "\n")


if __name__ == "__main__":
    parser = argparse.ArgumentParser("simple_example")
    parser.add_argument("--url", type=str)
    parser.add_argument("--method", type=str, choices=["post", "POST", "get", "GET"])
    parser.add_argument("--payload", type=json.loads, default={}, required=False)
    parser.add_argument("--user_jwt_token", type=str, required=False)
    args = parser.parse_args()

    if args.method.lower() == "post" and not args.payload:
        raise ValueError("--payload needs to be present for POST requests")

    if args.method.lower() == "get":
        make_get_request(url=args.url, user_jwt_token=args.user_jwt_token)
    else:
        make_post_request(
            url=args.url, payload=args.payload, user_jwt_token=args.user_jwt_token
        )
